from scarletio.tools.asynchronous_interactive_console import (
    collect_package_local_variables, create_banner, create_exit_message, run_asynchronous_interactive_console
)

from .... import __package__ as PACKAGE_NAME


PACKAGE = __import__(PACKAGE_NAME)

def __main__():
    run_asynchronous_interactive_console(
        collect_package_local_variables(PACKAGE),
        banner = create_banner(PACKAGE),
        exit_message = create_exit_message(PACKAGE),
    )



__all__ = ('import_extension',)

from os.path import basename as get_file_name, splitext as split_file_name_and_extension

from scarletio import export, get_last_module_frame

from ..extension import EXTENSIONS
from ..extension_loader import EXTENSION_LOADER


@export
def import_extension(extension_name, *variable_names, **keyword_parameters):
    """
    Imports from the given extension.
    
    Parameters
    ----------
    extension_name : `str`
        the extension's name to import.
    
    *variable_names : `str`
        Variables to load from the extension.
        
        > If no variable is defined, the module is returned instead.

    **keyword_parameters : Keyword parameters
        Additional parameters to create the extension with.
    
    Returns
    -------
    imported_variables : `Any`, `list` of `Any`
        The imported variable(s).
    
    Raises
    ------
    ImportError
        - Cannot find the designated extension.
    RuntimeError
        - If called from non extension file.
    TypeError
        - If `extension_name` is not a string.
        - If `variables_to_import` contains a non string.
    ValueError
        - If `extension_name`'s syntax is incorrect.
        - If `variable_names` contains a non identifier.
    ExtensionError
        - Any exception raised by the other extension is funneled.
    """
    # Validate input types
    if not isinstance(extension_name, str):
        raise TypeError(
            f'extension_name` can be `str`, got {extension_name.__class__.__name__}; {extension_name!r}.'
        )
    
    for variable_name in variable_names:
        if not isinstance(variable_name, str):
            raise TypeError(
                f'`variable_names` can contain only `str`, got {extension_name.__class__.__name__}; '
                f'{extension_name!r}; variable_names={variable_names!r}.'
            )
        
        if not variable_name.isidentifier():
            raise ValueError
        
    
    # Validate `extension_name` syntax.
    if not extension_name:
        raise ValueError(
            f'`extension_name`\s syntax is incorrect; Cannot be empty string.'
        )
    
    extension_name_parts = extension_name.split('.')
    expects_only_identifiers = False
    
    for extension_name_part in extension_name_parts:
        if not expects_only_identifiers:
            if not extension_name_part:
                continue
            
            expects_only_identifiers = True
        
        if not extension_name_part:
            raise ValueError(
                f'`extension_name`\s syntax is incorrect; two dot character cannot follow an identifier; got: '
                f'{extension_name!r}.'
            )
        
        if not extension_name_part.isidentifier():
            raise ValueError(
                f'`extension_name`\s syntax is incorrect; {extension_name_part!r} is not an identifier; got: '
                f'{extension_name!r}.'
            )
    
    if not extension_name_parts[-1]:
        raise ValueError(
            f'`extension_name`\s syntax is incorrect; Cannot end with dot character; got: {extension_name!r}.'
        )
    
    
    frame = get_last_module_frame()
    if (frame is None):
        spec = None
    else:
        spec = frame.f_globals.get('__spec__', None)
    
    if spec is None:
        raise RuntimeError(
            f'`import_extension` can only be called from an extension.'
        )
    
    local_name = spec.name
    
    is_init_file = (split_file_name_and_extension(get_file_name(spec.origin))[0] == '__init__')
    
    empty_count = 0
    for extension_name_part in extension_name_parts:
        if extension_name_part:
            break
        
        empty_count += 1
        continue
    
    if empty_count:
        local_name_parts = local_name.split('.')
        
        if len(local_name_parts) < empty_count:
            raise ImportError(
                f'`{extension_name!r}`\'s scope out of parent root\'s scope.'
            )
        
        extension_name_parts = extension_name_parts[empty_count:]
        
        empty_count = empty_count - is_init_file
        if empty_count:
            local_name_parts = local_name_parts[:-empty_count]
        
        built_name = '.'.join(local_name_parts+extension_name_parts)
        
    else:
        built_name = extension_name
    
    extension = EXTENSION_LOADER.load_extension(built_name, **keyword_parameters)
    
    current_extension = EXTENSIONS.get(local_name, None)
    if (current_extension is not None):
        extension.add_child_extension(current_extension)
        current_extension.add_parent_extension(extension)
    
    variable_names_length = len(variable_names)
    module = extension._spec.get_module_proxy()
    if variable_names_length == 0:
        return module
    
    variables = []
    
    for variable_name in variable_names:
        try:
            variable = getattr(module, variable_name)
        except AttributeError:
            raise ImportError(
                f'Cannot import `{variable_name}` from `{extension.name}`.'
            ) from None
        
        variables.append(variable)
    
    if variable_names_length == 1:
        return variables[0]
    
    return variables



__all__ = ()

from ... import Command

from scarletio.tools.asynchronous_interactive_console import AWAIT_NOTE

Command(
    'default',
    'help',
    'help',
    ['h'],
    'h | help *command*',
    'Either lists the available command, or shows the command\'s usage.',
)

Command(
    'default',
    'interpreter',
    'interpreter',
    ['i'],
    'i | interpreter',
    f'Runs asynchronous python interpreter through scarletio.\n{AWAIT_NOTE}',
)

Command(
    'default',
    'version',
    'version',
    ['v'],
    'v | version',
    'Prints out hata\'s version.'
)



import sys
from os import getcwd as get_current_work_directory
from os.path import (
    dirname as get_directory_name, expanduser as get_user_home_directory, join as join_paths,
    normpath as normalize_path, realpath as get_real_path
)


try:
    from . import __package__ as PACKAGE_NAME
except ImportError:
    # If we have hata not setupped
    PACKAGE_NAME = sys.path[0]
    
    sys.path.append(
        normalize_path(
            join_paths(
                get_directory_name(
                    get_real_path(
                        join_paths(
                            get_current_work_directory(),
                            get_user_home_directory(__file__),
                        )
                    )
                ),
                '..',
            )
        )
    )


PACKAGE = __import__(PACKAGE_NAME)

MAIN = __import__(f'{PACKAGE_NAME}.main').main

COMMAND_NAME_TO_COMMAND = MAIN.COMMAND_NAME_TO_COMMAND
find_commands = MAIN.find_commands

find_commands()

SYSTEM_DEFAULT_PARAMETER = 'i'

def command_not_found():
    from scarletio import get_short_executable
    
    output_parts = ['No command is added for: ']
    
    system_parameter = sys.argv
    
    length = len(system_parameter)
    if length > 2:
        index = 1
        
        while True:
            output_parts.append(repr(system_parameter[index]))
            index += 1
            if index == length:
                break
            
            output_parts.append(', ')
            continue
    else:
        output_parts.append('-')
    
    output_parts.append('\n')
    output_parts.append('Try using "$ ')
    output_parts.append(get_short_executable())
    output_parts.append(' ')
    output_parts.append(PACKAGE_NAME)
    output_parts.append(' help" for more information\n.')
    
    output = ''.join(output_parts)
    
    sys.stderr.write(output)


def __main__():
    system_parameters = sys.argv
    if len(system_parameters) < 2:
        system_parameter = SYSTEM_DEFAULT_PARAMETER
    else:
        system_parameter = system_parameters[1].lower()
    
    try:
        command = COMMAND_NAME_TO_COMMAND[system_parameter]
    except KeyError:
        command_function = command_not_found
    
    else:
        command_function = command.get_command_function()
    
    return command_function()


if __name__ == '__main__':
    __main__()



__all__ = ('find_commands',)

from os import listdir as list_directory
from os.path import isdir as is_directory, join as join_paths

from .constants import COMMANDS, COMMAND_FOLDER, COMMAND_IMPORT_ROUTE, COMMAND_NAME_TO_COMMAND


def find_commands():
    """
    Looks up the local commands.
    """
    COMMANDS.clear()
    COMMAND_NAME_TO_COMMAND.clear()
    
    for file_name in list_directory(COMMAND_FOLDER):
        path = join_paths(COMMAND_FOLDER, file_name)
        if is_directory(path):
            __import__('.'.join([*COMMAND_IMPORT_ROUTE, file_name]))



__all__ = ('EXTENSION_LOADER', 'ExtensionLoader', )

from functools import partial as partial_func

from scarletio import (
    HybridValueDictionary, RichAttributeErrorBaseType, Task, alchemy_incendiary, export, is_coroutine_function,
    render_exception_into, run_coroutine, shield
)

from ...discord.core import KOKORO

from .constants import EXTENSION_STATE_LOADED, EXTENSIONS
from .exceptions import ExtensionError
from .extension import Extension, __file__ as EXTENSION_LOADER_EXTENSION_FILE_PATH
from .helpers import (
    PROTECTED_NAMES, _build_extension_tree, _get_extension_name_and_path, _get_path_extension_name,
    _iter_extension_names_and_paths, _validate_entry_or_exit, validate_extension_parameters
)


EXTENSION_LOADER_EXTENSION_LOADER_FILE_PATH = __file__


def _render_exception_message(exception, header):
    """
    Renders an exception' traceback.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to render.
    header : `list` of `str`
        Header of the rendered traceback.
    
    Returns
    -------
    message : `str`
        The `exception`'s rendered traceback.
    
    Notes
    -----
    This function should run in an executor.
    """
    return ''.join(render_exception_into(exception, header, filter=_ignore_module_import_frames))


async def _render_exception_message_async(exception, header):
    """
    Returns the exception message inside of an executor.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to render.
    header : `list` of `str`
        Header of the rendered traceback.
    
    Returns
    -------
    message : `str`
        The `exception`'s rendered traceback.
    """
    return await KOKORO.run_in_executor(alchemy_incendiary(_render_exception_message, (exception, header)))


def _try_get_extension(extension_name, extension_path):
    """
    Tries to get extension for the given name or path.
    
    Parameters
    ----------
    extension_name : `None`, `str`
        Extension's  name.
    extension_path : `str`
        Path of the extension file.
    
    Returns
    -------
    extension : `None`, ``Extension``
        The found extension if any.
    """
    try:
        return EXTENSION_LOADER._extensions_by_name[extension_path]
    except KeyError:
        pass
    
    if extension_name is None:
        extension_name = _get_path_extension_name(extension_path)
    
    try:
        return EXTENSION_LOADER._extensions_by_name[extension_name]
    except KeyError:
        pass


async def _get_extensions(name, deep):
    """
    Gets the extensions with the given name.
    
    Parameters
    ----------
    name : `str`, `iterable` of `str`
        Extension by name to find.
    deep : `bool`
        Whether the extension with all of it's parent and with their child should be returned.
    
    Returns
    -------
    extensions : `list` of ``Extension``
        The found extensions.
    
    Raises
    ------
    ExtensionError
        If an extension could not be found.
    """
    try:
        extensions = set()
        
        extension_names_and_paths = set(_iter_extension_names_and_paths(name))
        for extension_name, extension_path in extension_names_and_paths:
            extension = _try_get_extension(extension_name, extension_path)
            if extension is None:
                raise ExtensionError(f'No extension was added with name: `{extension_name}`.')
            
            extensions.add(extension)
            continue
        
        if not extensions:
            raise ExtensionError(
                f'No extensions found with the given name: {name!r}.'
            )
        
        return _build_extension_tree(extensions, deep)
    
    except GeneratorExit:
        raise
    
    except ExtensionError:
        raise
    
    except BaseException as err:
        message = await _render_exception_message_async(
            err,
            [
                'Exception occurred meanwhile looking up extensions for name: `',
                repr(name),
                '`.\n',
            ]
        )
        
        raise ExtensionError(message)


def _ignore_module_import_frames(file_name, name, line_number, line):
    """
    Ignores import frames of extension loading.
    
    Parameters
    ----------
    file_name : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line_number : `int`
        The line's index where the exception occurred.
    line : `str`
        The frame's respective stripped line.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    if file_name.startswith('<') and file_name.endswith('>'):
        should_show_frame = False
    
    elif file_name == EXTENSION_LOADER_EXTENSION_FILE_PATH:
        if name == '_load':
            if line == 'loaded = self._load_module()':
                should_show_frame = False
        
        elif name == '_load_module':
            if line == 'spec.loader.exec_module(module)':
                should_show_frame = False
    
    elif file_name == EXTENSION_LOADER_EXTENSION_LOADER_FILE_PATH:
        if name == '_extension_loader_task':
            if line in (
                'module = await KOKORO.run_in_executor(extension._load)',
                'await entry_point(module)',
                'entry_point(module)',
            ):
                should_show_frame = False
        
        elif name == '_extension_unloader_task':
            if line in (
                'await exit_point(module)',
                'exit_point(module)',
            ):
                should_show_frame = False
    
    return should_show_frame


def _pop_unloader_task_callback(extension, task):
    """
    Removes the unloader task of the extension loader for the given extension.
    
    Parameters
    ----------
    extension : ``Extension``
        The extension to remove it's task of.
    task : ``Task``
        The finished task.
    """
    try:
        del EXTENSION_LOADER._unloader_tasks[extension]
    except KeyError:
        pass


def _pop_loader_task_callback(extension, task):
    """
    Removes the loader task of the extension loader for the given extension.
    
    Parameters
    ----------
    extension : ``Extension``
        The extension to remove it's task of.
    task : ``Task``
        The finished task.
    """
    try:
        del EXTENSION_LOADER._loader_tasks[extension]
    except KeyError:
        pass


def _run_maybe_blocking(coroutine, blocking):
    """
    Starts the given coroutine inside of a task. Returns a wrapped task, or it's result depending from where it was
    called.
    
    Parameters
    ----------
    coroutine : `CoroutineType`
        The coroutine to run.
    blocking : `bool`
        Whether the operation should be blocking when called from a non-async thread.
    
    Returns
    -------
    result : `None`, ``Task``, ``FutureAsyncWrapper``
    """
    if blocking:
        return run_coroutine(coroutine, KOKORO)
    else:
        return KOKORO.create_task_thread_safe(coroutine)


class ExtensionLoader(RichAttributeErrorBaseType):
    """
    There are some cases when you probably want to change some functional part of your client in runtime. Load,
    unload or reload code. Hata provides an easy to use (that's what she said) solution to solve this issue.
    
    It is called extension loader is an extension of hata. It is separated from `commands` extension, but it does not
    mean they do not go well together. But what more, extension loader was made to complement it.
    
    Usage
    -----
    The ``ExtensionLoader`` class is instanced already as `EXTENSION_LOADER` and that can be imported as well from
    `extension_loader.py`. Instancing the class again will yield the same object.
    
    Because hata can have more clients, we needed a special extension loader what can handle using more clients at any
    file, so the choice ended up on a  really interesting idea: assign variables to a module before it is (re)loaded.
    
    To show this is not blackmagic, here is an example:
    
    **main.py**
    
    ```py
    from hata.ext.extension_loader import EXTENSION_LOADER
    from datetime import datetime
    
    cake = 'cake'
    now = datetime.utcnow()
    
    EXTENSION_LOADER.add_default_variables(cake=cake, now=now)
    EXTENSION_LOADER.load_extension('extension')
    ```
    
    **extension.py**
    
    ```py
    print(cake, now)
    ```
    
    Make sure you start `main.py` with interactive mode. If you never did it, just use the `-i` option like:
    
    ```
    $ python3 -i main.py
    ```
    
    Or on windows:
    
    ```
    $ python -i main.py
    ```
    
    After you ran `main.py` you should see the following (the date should be different tho):
    
    ```
    cake 2020-03-14 09:40:41.587673
    ```
    
    Because now we have the interpreter, you can change the variables.
    
    ```py
    >>> EXTENSION_LOADER.add_default_variables(cake='cakes taste good, and now is:')
    >>> EXTENSION_LOADER.reload_all()
    ```
    
    And a different text is printed out:
    
    ```
    cakes taste good, and now is: 2020-03-14 09:40:41.587673
    ```
    
    Now lets edit `extension.py`.
    
    ```py
    cake = cake.split()
    print(*cake, now, sep='\\n')
    ```
    
    And reload the extension:
    
    ```py
    >>> EXTENSION_LOADER.reload_all()
    ```
    
    The printed text really changed again:
    
    ```
    cakes
    taste
    good,
    and
    now
    is:
    2020-03-14 09:40:41.587673
    ```
    
    If you remove default variables and the extension file still uses them, you get an ``ExtensionError``:
    
    ```py
    >>> EXTENSION_LOADER.remove_default_variables('cake')
    >>> EXTENSION_LOADER.reload_all()
    ```
    
    ```
    Traceback (most recent call last):
      File "<pyshell#13>", line 1, in <module>
        EXTENSION_LOADER.reload_all()
      File ".../hata/ext/extension_loader/extension_loader.py", line 652, in reload_all
        task.sync_wrap().wait()
      File ".../scarletio/core/traps/future_sync_wrapper.py", line 823, in wait
        return self.result()
      File ".../scarletio/core/traps/future_sync_wrapper.py", line 723, in result
        raise exception
      File ".../scarletio/core/traps/task.py", line 1602, in _step
        result = coroutine.throw(exception)
      File ".../hata/ext/extension_loader/extension_loader.py", line 670, in _reload_all_task
        raise ExtensionError(error_messages) from None
    hata.ext.extension_loader.ExtensionError: ExtensionError (1):
    Exception occurred meanwhile loading an extension: `extension`.
    
    Traceback (most recent call last):
      File ".../extension.py", line 1, in <module>
        cake = cake.split()
    NameError("name 'cake' is not defined")
    ```
    
    Adding Extensions
    -----------------
    Extensions can be added with the `.add` method.
    
    ```py
    EXTENSION_LOADER.add('cute_commands')
    ```
    
    Or more extension can be added as well by passing an iterable:
    
    ```py
    EXTENSION_LOADER.add(['cute_commands', 'nice_commands'])
    ```
    
    If an extension's file is not found, then `.add` will raise  `ModuleNotFoundError`. If the passed parameter is not
    `str` or not `iterable` of `str`, `TypeError` is raised.
    
    Loading
    -------
    Extensions can be loaded by their name:
    
    ```py
    EXTENSION_LOADER.load('cute_commands')
    ```
    
    All extension can be loaded by using:
    
    ```py
    EXTENSION_LOADER.load_all()
    ```
    
    Or extension can be added and loaded at the same time as well:
    
    ```py
    EXTENSION_LOADER.load_extension('cute_commands')
    ```
    
    `.load_extension` method supports all the keyword parameters as `.add`.
    
    ##### Passing variables to extensions
    
    You can pass variables to extensions with the `.add_default_variables` method:
    
    ```py
    EXTENSION_LOADER.add_default_variables(cake=cake, now=now)
    ```
    
    Adding or removing variables wont change the already loaded extensions' state, those needs to be reloaded to see
    them.
    
    Or pass variables to just specific extensions:
    
    ```py
    EXTENSION_LOADER.add('cute_commands', cake=cake, now=now)
    ```
    
    You can specify if the extension should use just it's own variables and ignore the default ones too:
    
    ```py
    EXTENSION_LOADER.add('cute_commands', extend_default_variables=False, cake=cake, now=now)
    ```
    
    Every variable added is stored in an optionally weak value dictionary, but you are able remove the added variables as well:
    
    ```py
    EXTENSION_LOADER.remove_default_variables('cake', 'now')
    ```
    
    The extensions can be accessed by their name as well, then you can use their properties to modify them.
    
    ```py
    EXTENSION_LOADER.get_extension('cute_commands').remove_default_variables('cake')
    ```
    
    Unloading And Exit Point
    ------------------------
    You can unload extension on the same way as loading them.
    
    ```py
    EXTENSION_LOADER.unload('cute_commands')
    ```
    
    Or unload all:
    
    ```py
    EXTENSION_LOADER.unload_all()
    ```
    
    When unloading an extension, the extension loader will search a function at the extension, what we call
    `exit_point` and will run it. By default it looks for a variable named `teardown`. `exit_point` acts on the same
    way as the `entry_point`, so it can be modified for looking for other name to defining and passing a callable
    (can be async again).
    
    Can be set almost on the same way as well:
    
    ```py
    EXTENSION_LOADER.default_exit_point = 'exit'
    ```
    
    Or:
    
    ```py
    EXTENSION_LOADER.add('cute_commands', exit_point='exit')
    ```
    
    There are also methods for reloading: `.reload(name)` and `.reload_all()`
    
    Removing Extensions
    -------------------
    
    Unloaded extensions can be removed from the extension loader by using the `.remove` method:
    
    ```py
    EXTENSION_LOADER.remove('cute_commands')
    ```
    
    Or more extension with an iterable:
    
    ```py
    EXTENSION_LOADER.remove(['cute_commands', 'nice_commands'])
    ```
    
    Removing loaded extension will yield `RuntimeError`.
    
    Threading Model
    ---------------
    When you call the different methods of the extension loader, they ll run on the ``Clients`` thread, what is named
    `KOKORO` internally.
    
    These methods are:
    - ``.load_extension``
    - ``.load``
    - ``.load_all``
    - ``.unload``
    - ``.unload_all``
    - ``.reload``
    - ``.reload_all``
    
    Meanwhile loading and executing the extension's code the thread is switched to an executor, so blocking tasks can
    be executed easily. The exception under this rule are the `entry_point`-s and the `exit_point`-s, which always run
    on the same thread as the clients, that is why they can be async as well.
    
    These methods also act differently depending from which thread they were called from. Whenever they are called from
    the client's thread, a ``Task`` is returned what can be `awaited`. If called from other ``EventThread``, then the
    task is async wrapped and that is returned. When calling from any other thread (like the main thread for example),
    the task is sync wrapped and the thread is blocked till the extension's loading is finished.
    
    Attributes
    ----------
    _default_entry_point : `None`, `str`, `callable`
        Internal slot for the ``.default_entry_point`` property.
    _default_exit_point : `None`, `str`, `callable`
        Internal slot for the ``.default_exit_point`` property.
    _default_variables : `None`, `HybridValueDictionary` of (`str`, `Any`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If it would be set as empty, then it is set as `None` instead.
    _done_callbacks : `None`, `list` of `callable`
        Callbacks, which run when loading / unloading the extensions is done.
        
        > These callbacks are only called once, and then cleared out.
    _execute_counter : `int`
        Whether the extension loader is executing an extension.
    _extensions_by_name : `dict` of (`str`, ``Extension``) items
        A dictionary of the added extensions to the extension loader in `extension-name` - ``Extension`` relation.
    _loader_tasks : `dict` of (``Extension``, ``Task``) items
        Active extension loading tasks.
    _unloader_tasks : `dict` of (``Extension``, ``Task``) items
        Active extension unloading tasks.
    
    Class Attributes
    ---------------
    _instance : `None`, ``ExtensionLoader``
        The already created instance of the ``ExtensionLoader`` if there is any.
    """
    __slots__ = (
        '_default_entry_point', '_default_exit_point', '_default_variables', '_done_callbacks', '_execute_counter',
        '_extensions_by_name', '_loader_tasks', '_unloader_tasks'
    )
    
    _instance = None
    
    def __new__(cls):
        """
        Creates an ``ExtensionLoader``. If the `ExtensionLoader` was instanced already, then returns that
        instead.
        
        Returns
        -------
        self : ``ExtensionLoader``
        """
        self = cls._instance
        if self is None:
            self = object.__new__(cls)
            
            self._default_entry_point = 'setup'
            self._default_exit_point = 'teardown'
            self._done_callbacks = None
            self._extensions_by_name = {}
            self._default_variables = HybridValueDictionary()
            self._execute_counter = 0
            self._loader_tasks = {}
            self._unloader_tasks = {}
            
            cls._instance = self
        
        return self
    
    
    @property
    def default_entry_point(self):
        """
        Get-set-del descriptor for modifying the extension loader's default entry point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._default_entry_point
    
    @default_entry_point.setter
    def default_entry_point(self, default_entry_point):
        if not _validate_entry_or_exit(default_entry_point):
            raise TypeError(
                f'`{self.__class__.__name__}.default_entry_point` can be `None`, `str`, `callable`, got '
                f'{default_entry_point.__class__.__name__}; {default_entry_point!r}.'
            )
        
        self._default_entry_point = default_entry_point
    
    @default_entry_point.deleter
    def default_entry_point(self):
        self._default_entry_point = None
    
    
    @property
    def default_exit_point(self):
        """
        Get-set-del descriptor for modifying the extension loader's default exit point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._default_exit_point
    
    
    @default_exit_point.setter
    def default_exit_point(self, default_exit_point):
        if not _validate_entry_or_exit(default_exit_point):
            raise TypeError(
                f'`{self.__class__.__name__}.default_exit_point` can be `None`, `str`, `callable`, got '
                f'{default_exit_point.__class__.__name__}; {default_exit_point!r}.'
            )
        
        self._default_exit_point = default_exit_point
    
    
    @default_exit_point.deleter
    def default_exit_point(self):
        self._default_exit_point = None
    
    
    def add_default_variables(self, **variables):
        """
        Adds default variables to the extension loader.
        
        Parameters
        ----------
        **variables : Keyword Parameters
            Variables to assigned to an extension's module before it is loaded.
        
        Raises
        ------
        ValueError
             If a variable name is would be used, what is `module` attribute.
        """
        default_variables = self._default_variables
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(
                    f'The passed {key!r} is a protected variable name of module type.'
                )
            
            default_variables[key] = value
    
    
    def remove_default_variables(self, *names):
        """
        Removes the mentioned default variables of the extension loader.
        
        If a variable with a specified name is not found, no error is raised.
        
        Parameters
        ----------
        *names : `str`
            Default variable names.
        """
        default_variables = self._default_variables
        for name in names:
            try:
                del default_variables[name]
            except KeyError:
                pass
    
    
    def clear_default_variables(self):
        """
        Removes all the default variables of the extension loader.
        """
        self._default_variables.clear()
    
    
    def add(self, name, *args, **kwargs):
        """
        Adds an extension to the extension loader.
        
        If the extension already exists, returns that one.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name to load.
        *args : Parameters
            Additional parameters to create the extension with.
        **kwargs : Keyword parameters
            Additional parameters to create the extension with.
        
        Other Parameters
        ----------------
        entry_point : `None`, `str`, `callable`, Optional
            Extension specific entry point, to use over the extension loader's default.
        exit_point : `None`, `str`, `callable`, Optional
            Extension specific exit point, to use over the extension loader's default.
        locked : `bool`, Optional
            Whether the given extension(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`, Optional
            Whether snapshot feature should be used.
        **variables : Keyword parameters
            Variables to assign to an extension(s)'s module before they are loaded.
            
        Raises
        ------
        ImportError
            If the given name do not refers to any loadable file.
        TypeError
            - If `entry_point` was not given as `None`, `str`, `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `exit_point` was not given as `None`, `str`, `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str`, `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        """
        extension_names_and_paths = set(_iter_extension_names_and_paths(name, register_directories_as_roots=True))
        entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference, default_variables = \
            validate_extension_parameters(*args, **kwargs)
        
        for extension_name, extension_path in extension_names_and_paths:
            self._add(
                extension_name, extension_path, entry_point, exit_point, extend_default_variables, locked,
                take_snapshot_difference, default_variables
            )
    
    
    def _add(self, extension_name, extension_path, entry_point, exit_point, extend_default_variables, locked,
            take_snapshot_difference, default_variables):
        """
        Adds an extension to the extension loader.
        
        If the extension already exists, returns that one.
        
        Parameters
        ----------
        extension_name : `None`, `str`
            The extension's name to add.
        extension_path : `None`, `str`
            Path of the extension to add.
        entry_point : `None`, `str`, `callable`
            Extension specific entry point, to use over the extension loader's default.
        exit_point : `None`, `str`, `callable`
            Extension specific exit point, to use over the extension loader's default.
        extend_default_variables : `bool`
            Whether the extension should use the loader's default variables or just it's own.
        locked : `bool`
            Whether the given extension(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`
            Whether snapshot feature should be used.
        default_variables : `None`, `HybridValueDictionary` of (`str`, `Any`) items
            An optionally weak value dictionary to store objects for assigning them to modules before loading them.
            If would be empty, is set as `None` instead.
        
        Returns
        -------
        extension : ``Extension``
            The registered extension.
        """
        extension = Extension(
            extension_name, extension_path, entry_point, exit_point, extend_default_variables,
            locked, take_snapshot_difference, default_variables
        )
        
        self._extensions_by_name[extension.name] = extension
        self._extensions_by_name[extension.file_name] = extension
        
        short_name = extension.short_name
        if (short_name is not None):
            self._extensions_by_name.setdefault(extension.short_name, extension)
        
        return extension
    
    
    def remove(self, name):
        """
        Removes one or more extensions from the extension loader.
        
        If any of the extensions is not found, no errors will be raised.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension(s)'s name(s) to remove.
        
        Raises
        ------
        TypeError
            If `name` was not given as `str`, `iterable` of `str`.
        RuntimeError
            If a loaded extension would be removed.
        """
        extension_names_and_paths = set(_iter_extension_names_and_paths(name))
        
        for extension_name, extension_path in extension_names_and_paths:
            if (extension_name is None):
                extension_name = _get_path_extension_name(extension_path)
            
            try:
                extension = self._extensions_by_name[extension_name]
            except KeyError:
                continue
            
            if extension._state == EXTENSION_STATE_LOADED:
                raise RuntimeError(
                    f'Extension `{name!r}` can not be removed meanwhile it is loaded.'
                )
        
        for extension_name, extension_path in extension_names_and_paths:
            if (extension_name is None):
                extension_name = _get_path_extension_name(extension_path)
            
            try:
                extension = self._extensions_by_name[extension_name]
            except KeyError:
                continue
            
            extension._unlink()
    
    
    def load_extension(self, name, *parameters, blocking=True, **keyword_parameters):
        """
        Adds, then loads the extension.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name to load.
        
        *parameters : Parameters
            Additional parameters to create the extension with.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the extension with.
        
        Other Parameters
        ----------------
        entry_point : `None`, `str`, `callable`, Optional
            Extension specific entry point, to use over the extension loader's default.
        exit_point : `None`, `str`, `callable`, Optional
            Extension specific exit point, to use over the extension loader's default.
        locked : `bool`, Optional
            Whether the given extension(s) should not be affected by `.{}_all` methods.
        take_snapshot_difference : `bool`, Optional
            Whether snapshot feature should be used.
        **variables : Keyword parameters
            Variables to assign to an extension(s)'s module before they are loaded.
        
        Returns
        -------
        task : `set` of ``Extension``, ``Task`` -> ``Extension``, ``FutureAsyncWrapper`` -> ``Extension``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
            
            When finished returns the loaded extension.
        
        Raises
        ------
        ImportError
            If the given name do not refers to any loadable file.
        TypeError
            - If `entry_point` was not given as `None`, `str`, `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `exit_point` was not given as `None`, `str`, `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str`, `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        ExtensionError
            The extension failed to load correctly.
        """
        return _run_maybe_blocking(self._load_extension_task(name, parameters, keyword_parameters), blocking)
    
    
    async def _load_extension_task(self, name, parameters, keyword_parameters):
        """
        Adds, then loads the extension.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`
            The extension's name.
        parameters : `tuple` of `Any`
            Additional parameters to create the extension with.
        keyword_parameters : `dict` of (`str`, `Any`) items
            Additional parameters to create the extension with.
        
        Returns
        -------
        extension : ``Extension``
            The loaded extension.
        
        Raises
        ------
        ImportError
            If the given name do not refers to any loadable file.
        TypeError
            - If `entry_point` was not given as `None`, `str`, `callable`.
            - If `entry_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `exit_point` was not given as `None`, `str`, `callable`.
            - If `exit_point` was given as `callable`, but accepts less or more positional parameters, as would be
                given.
            - If `extend_default_variables` was not given as `bool`.
            - If `locked` was not given as `bool`.
            - If `name` was not given as `str`, `iterable` of `str`.
        ValueError
            If a variable name is would be used, what is `module` attribute.
        ExtensionError
            The extension failed to load correctly.
        """
        extension_name, extension_path = _get_extension_name_and_path(name)
        extension = _try_get_extension(extension_name, extension_path)
        
        entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference, default_variables = \
            validate_extension_parameters(*parameters, **keyword_parameters)
    
        if (extension is None):
            extension = self._add(
                extension_name, extension_path, entry_point, exit_point, extend_default_variables, locked,
                take_snapshot_difference, default_variables
            )
            
        else:
            if (default_variables is not None):
                extension.add_default_variables(**default_variables)
            
            if extension.is_loaded():
                return extension
        
        extension_error = await self._extension_loader(extension)
        if (extension_error is not None):
            raise extension_error
        
        return extension
    
    
    def load(self, name, *, blocking=True, deep=True):
        """
        Loads the extension with the given name. If the extension is already loaded, will do nothing.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name.
        deep : `bool` = `True`, Optional (Keyword only)
            Whether the extension with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        task : `list` of ``Extension``, ``Task`` -> `list` of ``Extension``,
                ``FutureAsyncWrapper`` -> `list` of ``Extension``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
            
            When finished returns the loaded extension.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Loading the extension failed.
        """
        return _run_maybe_blocking(self._load_task(name, deep), blocking)
    
    
    async def _load_task(self, name, deep):
        """
        Loads the given extensions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str'
            The extension(s) name to load.
        deep : `bool`
            Whether the extension with all of it's parent and with their child should be reloaded.

        Returns
        -------
        extensions : `list` of ``Extension``
            The loaded extensions.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Loading the extension failed.
        """
        extensions = await _get_extensions(name, deep)
        
        error_messages = None
        
        for extension in extensions:
            exception = await self._extension_loader(extension)
            if (exception is not None):
                if error_messages is None:
                    error_messages = []
                
                error_messages.append(exception.message)
        
        if (error_messages is not None):
            raise ExtensionError(error_messages) from None
        
        return extensions
    
    
    def unload(self, name, *, blocking=True, deep=True):
        """
        Unloads the extension with the given name.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name.
        
        deep : `bool` = `True`, Optional (Keyword only)
            Whether the extension with all of it's parent and with their child should be reloaded.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Returns
        -------
        task : `list` of ``Extension``, ``Task`` -> `list` of ``Extension``,
                ``FutureAsyncWrapper`` -> `list` of ``Extension``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            unloading is done. However if called from a sync thread, will block till the unloading is done.
            
            When finished returns the unloaded extension.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Unloading the extension failed.
        """
        return _run_maybe_blocking(self._unload_task(name, deep), blocking)
    
    
    async def _unload_task(self, name, deep):
        """
        Unloads the extension with the given name.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name.
        deep : `bool`
            Whether the extension with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        extensions : `list` of ``Extension``
            The unloaded extensions.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Unloading the extension failed.
        """
        extensions = await _get_extensions(name, deep)
        
        error_messages = None
        
        for extension in extensions:
            exception = await self._extension_unloader(extension)
            if (exception is not None):
                if error_messages is None:
                    error_messages = []
                
                error_messages.append(exception.message)
        
        if (error_messages is not None):
            raise ExtensionError(error_messages) from None
        
        return extensions
    
    
    def reload(self, name, *, blocking=True, deep=True):
        """
        Reloads the extension with the given name.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name.

        deep : `bool` = `True`, Optional (Keyword only)
            Whether the extension with all of it's parent and with their child should be reloaded.
        
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Returns
        -------
        task : `list` of ``Extension``, ``Task`` -> `list` of ``Extension``,
                ``FutureAsyncWrapper`` -> `list` of ``Extension``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            reloading is done. However if called from a sync thread, will block till the reloading is done.
            
            When finished returns the reloaded extensions.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Reloading the extension failed.
        """
        return _run_maybe_blocking(self._reload_task(name, deep), blocking)
    
    
    async def _reload_task(self, name, deep):
        """
        Reloads the extension with the given name.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`, `iterable` of `str`
            The extension's name.
        deep : `bool`
            Whether the extension with all of it's parent and with their child should be reloaded.
        
        Returns
        -------
        extensions : `list` of ``Extension``
            The reloaded extensions.
        
        Raises
        ------
        ExtensionError
            - No extension is added with the given name.
            - Reloading the extension failed.
        """
        extensions = await _get_extensions(name, deep)
        
        await self._check_for_syntax(extensions)
        
        error_messages = None
        
        for extension in extensions:
            exception = await self._extension_unloader(extension)
            if (exception is not None):
                if error_messages is None:
                    error_messages = {}
                
                error_messages[extension.name] = exception.message
        
        
        for extension in extensions:
            if (error_messages is not None) and (extension.name in error_messages):
                continue
            
            exception = await self._extension_loader(extension)
            if (exception is not None):
                if error_messages is None:
                    error_messages = {}
                
                error_messages[extension.name] = exception.message
        
        if (error_messages is not None):
            raise ExtensionError([*error_messages.values()]) from None
        
        return extensions
    
    
    def load_all(self, *, blocking=True):
        """
        Loads all the extension of the extension loader. If anything goes wrong, raises an ``ExtensionError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            loading is done. However if called from a sync thread, will block till the loading is done.
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        ExtensionError
            If any extension failed to load correctly.
        """
        return _run_maybe_blocking(self._load_all_task(), blocking)
    
    
    async def _load_all_task(self):
        """
        Loads all the extensions of the extension loader.
        
        Loads each extension one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised only at the end. If any of the extensions raises, will still try to unload the leftover
        ones.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            If any extension failed to load correctly.
        """
        error_messages = None
        
        for extension in tuple(EXTENSIONS.values()):
            if extension._locked:
                continue
            
            exception = await self._extension_loader(extension)
            if (exception is not None):
                if error_messages is None:
                    error_messages = []
                
                error_messages.append(exception.message)
        
        if (error_messages is not None):
            raise ExtensionError(error_messages) from None
    
    
    def unload_all(self, *, blocking=True):
        """
        Unloads all the extension of the extension loader. If anything goes wrong, raises an ``ExtensionError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            unloading is done. However if called from a sync thread, will block till the unloading is done.
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        ExtensionError
            If any extension failed to unload correctly.
        """
        return _run_maybe_blocking(self._unload_all_task(), blocking)
    
    
    async def _unload_all_task(self):
        """
        Unloads all the extensions of the extension loader.
        
        Unloads each extension one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised only at the end. If any of the extensions raises, will still try to unload the leftover
        ones.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            If any extension failed to unload correctly.
        """
        error_messages = None
        
        for extension in tuple(EXTENSIONS.values()):
            if extension._locked:
                continue
            
            exception = await self._extension_unloader(extension)
            if (exception is not None):
                if error_messages is None:
                    error_messages = []
                
                error_messages.append(exception.message)
            
        if (error_messages is not None):
            raise ExtensionError(error_messages) from None
    
    
    def reload_all(self, *, blocking=True):
        """
        Reloads all the extension of the extension loader. If anything goes wrong, raises an ``ExtensionError`` only
        at the end, with the exception(s).
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureAsyncWrapper``
            If the method is called from an ``EventThread``, then returns an awaitable, what will yield when the
            reloading is done. However if called from a sync thread, will block till the reloading is done.
        blocking : `bool` = `True`, Optional (Keyword only)
            Whether the operation should be blocking when called from a non-async thread.
        
        Raises
        ------
        ExtensionError
            If any extension failed to reload correctly.
        """
        return _run_maybe_blocking(self._reload_all_task(), blocking)
    
    
    async def _reload_all_task(self):
        """
        Reloads all the extensions of the extension loader.
        
        If an extension is not unloaded, will load it, if the extension is loaded, will unload, then load it.
        Reloads each extension one after the other. The raised exceptions' messages are collected into one exception,
        what will be raised at the end. If any of the extensions raises, will still try to reload the leftover ones.
        
        This method is a coroutine.
        
        Raises
        ------
        ExtensionError
            If any extension failed to reload correctly.
        """
        error_messages = None
        
        extensions = _build_extension_tree(
            [extension for extension in EXTENSIONS.values() if not extension._locked],
            False,
        )
        await self._check_for_syntax(extensions)
        
        for extension in extensions:
            exception = await self._extension_unloader(extension)
            if (exception is None):
                exception = await self._extension_loader(extension)
            
            if (exception is not None):
                if error_messages is None:
                    error_messages = []
                
                error_messages.append(exception.message)
        
        if (error_messages is not None):
            raise ExtensionError(error_messages) from None
    
    
    async def _extension_loader(self, extension):
        """
        Loads the given extension. This method synchronises extension loading.
        
        This method is a coroutine.
        
        Parameters
        ----------
        extension : ``Extension``
            The extension to unload.
        
        Returns
        -------
        exception : `None`, ``ExtensionError``
        """
        try:
            unloader_task = self._unloader_tasks[extension]
        except KeyError:
            pass
        else:
            try:
                exception = await shield(unloader_task, KOKORO)
            except:
                raise
            else:
                if (exception is not None):
                    return exception
            
            finally:
                unloader_task = None
            
        try:
            loader_task = self._loader_tasks[extension]
        except KeyError:
            loader_task = Task(self._extension_loader_task(extension), KOKORO)
            loader_task.add_done_callback(partial_func(_pop_loader_task_callback, extension))
            
            self._loader_tasks[extension] = loader_task
        
        try:
            return await shield(loader_task, KOKORO)
        finally:
            loader_task = None
    
    
    async def _extension_loader_task(self, extension):
        """
        Loads the extension. If the extension is loaded, will do nothing.
        
        Loading an exception can be separated to 4 parts:
        
        - Assign the default variables.
        - Load the module.
        - Find the entry point (if needed).
        - Ensure the entry point (if found).
        
        If any of these fails, an ``ExtensionError`` will be raised. If step 1 raises, then a traceback will be
        included as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        extension : ``Extension``
            The extension to load.
        
        Returns
        -------
        exception : `None`, ``ExtensionError``
        """
        self._execute_counter += 1
        try:
            try:
                # loading blocks, but unloading does not
                module = await KOKORO.run_in_executor(extension._load)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                message = await _render_exception_message_async(
                    err,
                    [
                        'Exception occurred meanwhile loading an extension: `',
                        extension.name,
                        '`.\n\n',
                    ],
                )
                
                return ExtensionError(message)
            
            if module is None:
                return # already loaded
            
            entry_point = extension._entry_point
            if entry_point is None:
                entry_point = self._default_entry_point
                if entry_point is None:
                    return
            
            if isinstance(entry_point, str):
                entry_point = getattr(module, entry_point, None)
                if entry_point is None:
                    return
            
            try:
                if is_coroutine_function(entry_point):
                    await entry_point(module)
                else:
                    entry_point(module)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                message = await _render_exception_message_async(
                    err,
                    [
                        'Exception occurred meanwhile entering an extension: `',
                        extension.name,
                        '`.\nAt entry_point:',
                        repr(entry_point),
                        '\n\n',
                    ],
                )
                
                return ExtensionError(message)
        finally:
            execute_counter = self._execute_counter - 1
            self._execute_counter = execute_counter
            if not execute_counter:
                self.call_done_callbacks()
    
    
    async def _extension_unloader(self, extension):
        """
        Unloads the given extension. This method synchronises extension unloading.
        
        This method is a coroutine.
        
        Parameters
        ----------
        extension : ``Extension``
            The extension to unload.
        
        Returns
        -------
        exception : `None`, ``ExtensionError``
        """
        try:
            loader_task = self._loader_tasks[extension]
        except KeyError:
            pass
        else:
            try:
                exception = await shield(loader_task, KOKORO)
            except:
                raise
            else:
                if (exception is not None):
                    return exception
            
            finally:
                loader_task = None
        
        try:
            unloader_task = self._unloader_tasks[extension]
        except KeyError:
            unloader_task = Task(self._extension_unloader_task(extension), KOKORO)
            unloader_task.add_done_callback(partial_func(_pop_unloader_task_callback, extension, ))
            
            self._unloader_tasks[extension] = unloader_task
        
        try:
            return await shield(unloader_task, KOKORO)
        finally:
            unloader_task = None

    
    async def _extension_unloader_task(self, extension):
        """
        Unloads the extension. If the extension is not loaded, will do nothing.
        
        Loading an exception can be separated to 3 parts:
        
        - Find the exit point (if needed).
        - Ensure the exit point (if found).
        - Remove the default variables.
        
        If any of these fails, an ``ExtensionError`` will be raised. If step 2 raises, then a traceback will be
        included as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        extension : ``Extension``
            The extension to unload.
        
        Returns
        -------
        exception : `None`, ``ExtensionError``
        """
        self._execute_counter += 1
        try:
            # loading blocks, but unloading does not
            module = extension._unload()
            
            if module is None:
                return # not loaded
            
            try:
                exit_point = extension._exit_point
                if exit_point is None:
                    exit_point = self._default_exit_point
                    if exit_point is None:
                        return
                
                if isinstance(exit_point, str):
                    exit_point = getattr(module, exit_point, None)
                    if exit_point is None:
                        return
                
                try:
                    
                    if is_coroutine_function(exit_point):
                        await exit_point(module)
                    else:
                        exit_point(module)
                except GeneratorExit:
                    raise
                
                except BaseException as err:
                    message = await _render_exception_message_async(
                        err,
                        [
                            'Exception occurred meanwhile unloading an extension: `',
                            extension.name,
                            '`.\nAt exit_point:',
                            repr(exit_point),
                            '\n\n',
                        ],
                    )
                    
                    return ExtensionError(message)
            
            finally:
                extension._unassign_variables()
                
                keys = []
                module_globals = module.__dict__
                for key in module_globals:
                    if (not key.startswith('__')) and (not key.endswith('__')):
                        keys.append(key)
                
                for key in keys:
                    del module_globals[key]
        finally:
            execute_counter = self._execute_counter - 1
            self._execute_counter = execute_counter
            if not execute_counter:
                self.call_done_callbacks()
    
    
    def __repr__(self):
        """Returns the extension loader's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' extension count=',
            repr(len(EXTENSIONS)),
        ]
        
        entry_point = self._default_entry_point
        if (entry_point is not None):
            repr_parts.append(', default_entry_point=')
            repr_parts.append(repr(entry_point))
        
        exit_point = self._default_exit_point
        if (exit_point is not None):
            repr_parts.append(', default_exit_point=')
            repr_parts.append(repr(exit_point))
        
        default_variables = self._default_variables
        if default_variables:
            repr_parts.append(', default_variables=')
            repr_parts.append(repr(default_variables))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def is_processing_extension(self):
        """
        Returns whether the extension loader is processing an extension.
        
        Returns
        -------
        is_processing_extension : `bool`
        """
        if self._execute_counter:
            is_processing_extension = True
        else:
            is_processing_extension = False
        
        return is_processing_extension
    
    
    def get_extension(self, name):
        """
        Returns the extension loader's extension with the given name.
        
        Parameters
        ----------
        name : `str`
            An extension's name.
        
        Returns
        -------
        extension : ``Extension``, `None`.
            The matched extension if any.
        """
        return self._extensions_by_name.get(name, None)
    
    
    async def _check_for_syntax(self, extensions):
        """
        Checks whether the extensions can be reloaded.
        
        This function is a coroutine.
        
        Parameters
        ----------
        extensions : `list` of ``Extension``
            A list of extensions to check their syntax.
        
        Raises
        ------
        SyntaxError
        """
        return await KOKORO.run_in_executor(alchemy_incendiary(self._check_for_syntax_blocking, (extensions,)))
    
    
    def _check_for_syntax_blocking(self, extensions):
        """
        Checks whether the extensions can be reloaded.
        
        This method is blocking and ran inside of an executor by ``._check_for_syntax``.
        
        Parameters
        ----------
        extensions : `list` of ``Extension``
            A list of extensions to check their syntax.
        
        Raises
        ------
        SyntaxError
        """
        for extension in extensions:
            extension._check_for_syntax()
    
    
    def add_done_callback(self, callback):
        """
        Adds a done callback to be called when the loading / unloading extensions is finished.
        
        These callbacks are only called once, and then cleared out.
        
        Parameters
        ----------
        callback : `callable`
        
        Returns
        -------
        added : `bool`
            Whether the callback was added.
        """
        done_callbacks = self._done_callbacks
        if (done_callbacks is None):
            done_callbacks = []
            self._done_callbacks = done_callbacks
        
        done_callbacks.append(callback)
        return True
    
    
    def add_done_callback_unique(self, callback):
        """
        Adds a done callback to be called when the loading / unloading extensions is finished.
        
        > Ff the callback is already added, does nothing.
        
        These callbacks are only called once, and then cleared out.
        
        Parameters
        ----------
        callback : `callable`
        
        Returns
        -------
        added : `bool`
            Whether the callback was added.
        """
        done_callbacks = self._done_callbacks
        if (done_callbacks is None):
            done_callbacks = []
            self._done_callbacks = done_callbacks
        
        else:
            if (callback in done_callbacks):
                return False
        
        done_callbacks.append(callback)
        return True
    
    
    def call_done_callbacks(self):
        """
        Calls done callbacks of the extension loader.
        """
        done_callbacks = self._done_callbacks
        if (done_callbacks is not None):
            self._done_callbacks = done_callbacks
            
            for done_callback in done_callbacks:
                KOKORO.call_soon(done_callback)


EXTENSION_LOADER = ExtensionLoader()
export(EXTENSION_LOADER, 'EXTENSION_LOADER')



__all__ = ()

from scarletio import include

from ...core import STAGES


AuditLogEvent = include('AuditLogEvent')
Invite = include('Invite')

def target_converter_none(entry):
    return None


def target_converter_guild(entry):
    parent = entry.parent
    if (parent is not None):
        return parent.guild


def target_converter_channel(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.channels.get(target_id, None)


def target_converter_user(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.users.get(target_id, None)


def target_converter_role(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.guild.roles.get(target_id, None)


def target_converter_invite(entry):
    # every other data is at # change
    for change in entry.changes:
        if change.attribute_name != 'code':
            continue
        
        if entry.type is AuditLogEvent.invite_delete:
            code = change.before
        else:
            code = change.after
        break
    
    else:
        code = '' # malformed ?
    
    
    parent = entry.parent
    if (parent is None):
        guild = None
    else:
        guild = parent.guild
    
    return Invite.precreate(code, guild=guild)


def target_converter_webhook(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.webhooks.get(target_id, None)


def target_converter_emoji(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.guild.emojis.get(target_id, None)


def target_converter_integration(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.integrations.get(target_id, None)


def target_converter_stage(entry):
    target_id = entry.target_id
    if target_id:
        return STAGES.get(target_id, None)


def target_converter_sticker(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.guild.stickers.get(target_id, None)


def target_converter_scheduled_event(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.scheduled_events.get(target_id, None)


def target_converter_thread(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.threads(target_id, None)


def target_converter_application_command(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.application_commands(target_id, None)


def target_converter_auto_moderation(entry):
    return None



__all__ = ('Command',)

import warnings

from scarletio import RichAttributeErrorBaseType

from .constants import COMMAND_IMPORT_ROUTE, COMMANDS, COMMAND_NAME_TO_COMMAND


class Command(RichAttributeErrorBaseType):
    """
    Command class for command lookup.
    
    Attributes
    ----------
    alters : `None` or `frozenset` of `str`
        Alternative names of the command.
    description : `str`
        The command's description.
    file_name : `str`
        The command's file's name.
    diretcory_name : `str`
        The command's diretcory_name.
    name : `str`
        The command's name.
    usage : `str`
        The command's usage.
    """
    __slots__ = ('alters', 'description', 'file_name', 'diretcory_name', 'name', 'usage')
    
    def __new__(cls, diretcory_name, file_name, name, alters, usage, description):
        """
        Parameters
        ----------
        diretcory_name : `str`
            The command's diretcory_name.
        file_name : `str`
            The command's file's name.
        name : `str`
            The command's name.
        alters : `None` or `list` of `str`
            Alternative names of the command.
        usage : `str`
            The command's usage.
        description : `str`
            The command's description.
        """
        if file_name.endswith('.py'):
            file_name = file_name[:-3]
        
        if (alters is not None):
            if alters:
                alters = frozenset(alters)
            else:
                alters = None
        
        self = object.__new__(cls)
        self.alters = alters
        self.description = description
        self.file_name = file_name
        self.diretcory_name = diretcory_name
        self.name = name
        self.usage = usage
        
        COMMANDS.add(self)
        
        try:
            old_command = COMMAND_NAME_TO_COMMAND[name]
        except KeyError:
            pass
        else:
            if old_command.name == name:
                warnings.warn(
                    RuntimeError,
                    (
                        f'Multiple commands with name: {name}\n'
                        f'- {old_command.diretcory_name}.{old_command.name}\n'
                        f'- {diretcory_name}.{name}'
                    ),
                )
        
        COMMAND_NAME_TO_COMMAND[name] = self
        if (alters is not None):
            for alter in alters:
                COMMAND_NAME_TO_COMMAND.setdefault(alter, self)
        
        return self
    
    
    def __hash__(self):
        """Returns the command's hash value."""
        hash_value = 0
        
        # alters
        alters = self.alters
        if (alters is not None):
            hash_value ^= hash(alters)
        
        # description
        hash_value ^= hash(self.description)
        
        # file_name
        file_name = self.file_name
        try:
            file_name_hash_value = hash(file_name)
        except (NotImplementedError, RuntimeError, TypeError):
            file_name_hash_value = object.__hash__(file_name)
        hash_value ^= file_name_hash_value
        
        # diretcory_name
        hash_value ^= hash(self.diretcory_name)
        
        # name
        hash_value ^= hash(self.name)
        
        # usage
        hash_value ^= hash(self.usage)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two commands are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # alters
        if self.alters != other.alters:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # file_name
        if self.file_name != other.file_name:
            return False
        
        # diretcory_name
        if self.diretcory_name != other.diretcory_name:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # usage
        if self.usage != other.usage:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the command's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' ')
        repr_parts.append(self.name)
        repr_parts.append(' from ')
        
        repr_parts.append(self.diretcory_name)
        repr_parts.append('.')
        repr_parts.append(self.file_name)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_command_function(self):
        """
        Returns the command function of the command.
        
        Returns
        -------
        command_function : `FunctionType`
        """
        access_path_parts = [*COMMAND_IMPORT_ROUTE, self.diretcory_name, self.file_name]
        module = __import__('.'.join(access_path_parts))
        
        for access_path_part in access_path_parts[1:]:
            module = getattr(module, access_path_part)
        
        return module.__main__



import sys
from math import floor, log10

from scarletio import get_short_executable

from .... import __package__ as PACKAGE_NAME

from ... import COMMANDS, COMMAND_NAME_TO_COMMAND


PACKAGE = __import__(PACKAGE_NAME)


def show_command(command_name):
    command = COMMAND_NAME_TO_COMMAND.get(command_name.lower(), None)
    
    output_parts = []
    if command is None:
        output_parts = []
        
        output_parts.append('No command is added for: ')
        output_parts.append(repr(command_name))
        output_parts.append('\n')
        output_parts.append('Try using "$ ')
        output_parts.append(get_short_executable())
        output_parts.append(' ')
        output_parts.append(PACKAGE_NAME)
        output_parts.append(' help" to list all available commands\n.')
    
    else:
        output_parts.append('Help for: "')
        output_parts.append(command.name)
        output_parts.append('"\n\nUsage: "$ ')
        output_parts.append(get_short_executable())
        output_parts.append(' ')
        output_parts.append(PACKAGE_NAME)
        output_parts.append(' ')
        output_parts.append(command.usage)
        output_parts.append('"\n\n')
        output_parts.append(command.description)
        output_parts.append('\n')
    
    output = ''.join(output_parts)
    sys.stderr.write(output)


def list_commands():
    output_parts = []
    output_parts.append('Available commands:\n\n')
    
    command_count = len(COMMANDS)
    index_adjust = floor(log10(command_count)) + 1
    
    for index, command in enumerate(COMMANDS, 1):
        command_name = command.name
        output_parts.append(str(index).rjust(index_adjust))
        output_parts.append('.: ')
        output_parts.append(command_name)
        output_parts.append('\n')
    
    output_parts.append('\n')
    output_parts.append('Use "$ ')
    output_parts.append(get_short_executable())
    output_parts.append(' ')
    output_parts.append(PACKAGE_NAME)
    output_parts.append(' help ')
    output_parts.append('*command*')
    output_parts.append('" for more information.\n')
    
    output = ''.join(output_parts)
    sys.stderr.write(output)


def __main__():
    system_parameters = sys.argv
    if len(system_parameters) < 3:
        command_parameter = None
    else:
        command_parameter = system_parameters[2].lower()
    
    if command_parameter is None:
        list_commands()
    else:
        show_command(command_parameter)



__all__ = ('bind',)

from scarletio import RichAttributeErrorBaseType, WeakKeyDictionary, copy_docs


def bind(bind_to, bind_with, name):
    """
    A cool hata features, that lets you bind object to existing one.
    
    Parameters
    ----------
    bind_to : `type`
        The type to bind to.
    bind_with : `type`
        The type to bind with.
    name : `str`
        The name of the binding.
    
    Raises
    ------
    TypeError
        - If `bind_to` is not a type.
        - If `bind_to`-s do not support weakreferencing.
        - If `bind_to` has already an attribute named `name`.
    
    Examples
    --------
    Generic binder:
    ```py
    from hata import ClientUserBase, bind
    
    class Inventory(object):
        def __init__(self, parent_self):
            self.user_id = parent_self.id
            self.inventory = []
        
        def add_item(self, name):
            self.inventory.append(name)
        
        def give_user(self, item, recipient):
            self.inventory.remove(item)
            recipient.inventory.add_item(item)
    
    bind(ClientUserBase, Inventory, 'inventory')
    
    # Usage:
    user.inventory.add_item('cake')
    ```
    
    Descriptor binder:
    ```py
    from hata import ClientUserBase, bind
    
    class ValueBinder(object):
        def __init__(self, parent_self):
            self.value = 10
        
        def __get__(self, instance, type_):
            return self.value
        
        def __set__(self, instance, value):
            self.value = value
    
    bind(ClientUserBase, ValueBinder, 'value')
    
    # Usage:
    user.inventory_size += 10
    ```
    Descriptor binders are familiar ot normal descriptors, except, they are created per object and not per class.
    
    Sometimes you want to clear extra binders with default values. For this you might want to define the `__bool__`
    method.
    ```py
    DEFAULT_VALUE = 10
    
    class ValueBinder(object):
        def __init__(self, parent_self):
            self.value = DEFAULT_VALUE
        
        def __get__(self, instance, type_):
            return self.value
        
        def __set__(self, instance, value):
            self.value = value
        
        def __bool__(self):
            return (self.value != DEFAULT_VALUE)
    
    
    bind(ClientUserBase, ValueBinder, 'value')
    
    # Usage
    ClientUserBase.value.clear()
    ```
    
    A more advanced concept is reloading modules, which use bindings. For this define `__getstate__` and the
    `__setstate__` magic methods.
    
    > Also defining `__bool__` is not required, but recommended.
    ```py
    class Inventory(object):
        def __init__(self, parent_self):
            self.user_id = parent_self.id
            self.inventory = []
        
        def add_item(self, name):
            self.inventory.append(name)
        
        def give_user(self, item, recipient):
            self.inventory.remove(item)
            recipient.inventory.add_item(item)
        
        def __getstate__(self):
            return (self.user_id, self.inventory)
        
        def __setstate(self, state):
            self.user_id, self.inventory = state
    
    bind(ClientUserBase, Inventory, 'inventory')
    ```
    """
    if not isinstance(bind_to, type):
        raise TypeError(
            f'`bind_to` can be `type`, got {bind_to.__class__.__name__}; {bind_to!r}.'
        )
    
    name_space = bind_to.__dict__
    if ('__weakref__' in name_space) or ('__slots__' not in name_space):
        raise TypeError(
            f'`bind_to`-s must support weakreferencing, got {bind_to!r}.'
        )
    
    if hasattr(bind_to, name):
        old_binder = getattr(bind_to, name)
        if not isinstance(old_binder, ObjectBinderBase):
            raise TypeError(
                f'`bind_to` already has an attribute named, as bind_to={bind_to!r}, name={name!r}.'
            )
    
    else:
        old_binder = None
    
    if (getattr(bind_with, '__get__', None) is not None):
        new_binder = DescriptorObjectBinder(name, bind_with)
    else:
        new_binder = GenericObjectBinder(name, bind_with)
    
    if (
        (old_binder is not None) and
        old_binder.supports_state_transfer and
        new_binder.supports_state_transfer and
        (old_binder.get_type_key() == new_binder.get_type_key())
    ):
        new_binder.set_states(old_binder.get_states(bind_to))
    
    setattr(bind_to, name, new_binder)


class ObjectBinderBase(RichAttributeErrorBaseType):
    """
    Object binder.
    
    Attributes
    ----------
    cache : ``WeakKeyDictionary``
        The binding relation mapping.
    name : `str`
        The filed's name.
    supports_clearing : `bool`
        Whether clearing the bound objects is supported.
    supports_state_transfer : `bool`
        Whether the bound objects support state transfer.
    type : `type`
        The binded class.
    """
    __slots__ = ('cache', 'name', 'supports_clearing', 'supports_state_transfer', 'type')
    
    def __new__(cls, name, type_):
        """
        Creates a new object binder instance.
        
        Parameters
        ----------
        name : `str`
            The filed's name.
        type_ : `type`
            The binded class.
        """
        bool_method = getattr(type_, '__bool__', None)
        if (bool_method is None):
            supports_clearing = False
        else:
            supports_clearing = True
        
        get_state_method = getattr(type_, '__getstate__', None)
        set_state_method = getattr(type_, '__setstate__', None)
        if (get_state_method is None) or (set_state_method is None):
            supports_state_transfer = False
        else:
            supports_state_transfer = True
        
        self = object.__new__(cls)
        self.cache = WeakKeyDictionary()
        self.name = name
        self.type = type_
        
        self.supports_clearing = supports_clearing
        self.supports_state_transfer = supports_state_transfer
        
        return self
    
    
    def __get__(self, instance, type_):
        """
        Gets the field's value.
        
        Parameters
        ----------
        instance : `Any`
            The instance itself.
        type_ : `type`
            The instance's type.
        
        Returns
        -------
        value : `self` / `Any`
            Returns itself if called from class.
        """
        if (instance is None):
            return self
    
    
    def __set__(self, instance, value):
        """
        Sets the value to the instance.
        
        Parameters
        ----------
        instance : `Any`
            The parent instance to set the value to.
        value : `Any`
            The value to set.
        
        Raises
        ------
        AttributeError
            Read only attribute.
        """
        raise AttributeError(f'Read only attribute: {self.name!r}.')
    
    
    def __delete__(self, instance):
        """
        Deletes the value of the instance.
        
        Parameters
        ----------
        instance : `Any`
            The parent instance to delete the attribute of.
        
        Raises
        ------
        AttributeError
            Cannot delete attribute.
        """
        raise AttributeError(f'Cannot delete attribute: {self.name!r}.')
    
    
    def __repr__(self):
        """Returns the object binder representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        repr_parts.append(', type=')
        repr_parts.append(repr(self.type))
        
        length = len(self.cache)
        if length:
            repr_parts.append(', length=')
            repr_parts.append(repr(length))
        
        supports_clearing = self.supports_clearing
        if supports_clearing:
            repr_parts.append(', supports_clearing=')
            repr_parts.append(repr(supports_clearing))
        
        supports_state_transfer = self.supports_state_transfer
        if supports_state_transfer:
            repr_parts.append(', supports_state_transfer=')
            repr_parts.append(repr(supports_state_transfer))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def clear(self):
        """
        Clears the bound objects.
        
        Only those objects are cleared, which return `False` from their `__bool__` method.
        
        Returns
        -------
        cleared : `int`
            The amount of objects freed.
        """
        if not self.supports_clearing:
            return 0
        
        cache = self.cache
        to_clear = []
        
        for key, value in cache.items():
            if not value:
                to_clear.append(key)
        
        for key in to_clear:
            del cache[key]
        
        return len(to_clear)
    
    
    def get_type_key(self):
        """
        Returns the type key of the object binder.
        
        It is used when transferring down states to match types.
        
        Returns
        -------
        type_key : `tuple` ((`None`, `str`), (`None`, `str`))
        """
        type_ = self.type
        return getattr(type_, '__module__', None), getattr(type_, '__name__', None)
    
    
    def iter_true_items(self):
        """
        Iterates over all non empty bound objects & keys.
        
        This method is an iterable generator.
        
        Yields
        ------
        key : `Any`
            Objects to bound to.
        value : ``.type``
            The bound object.
        """
        if self.supports_clearing:
            for key, value in self.cache.items():
                if value:
                    yield key, value
        
        else:
            yield from self.cache.items()
    
    
    def get_states(self, type_):
        """
        Gets the given object's states if applicable for the given `type_` instances.
        
        Returns
        -------
        states : `list` of `tuple` (`Any`, `Any`)
        """
        states = []
        if self.supports_state_transfer:
            get_state = self.type.__getstate__
            
            for key, value in self.iter_true_items():
                if not isinstance(key, type_):
                    continue
                
                state = get_state(value)
                states.append((key, state))
        
        return states
    
    
    def set_states(self, states):
        """
        Sets states of an another binder to this one.
        
        Parameters
        ----------
        states : `list` of `tuple` (`Any`, `Any`)
        """
        if self.supports_state_transfer:
            type_ = self.type
            set_state = type_.__setstate__
            cache = self.cache
            
            for key, state in states:
                value = object.__new__(type_)
                set_state(value, state)
                
                cache[key] = value


@copy_docs(ObjectBinderBase)
class GenericObjectBinder(ObjectBinderBase):
    __slots__ = ()
    
    @copy_docs(ObjectBinderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return bind_object


class DescriptorObjectBinder(ObjectBinderBase):
    """
    Object binder for descriptors-likes.
    
    Attributes
    ----------
    cache : ``WeakKeyDictionary``
        The binding relation mapping.
    name : `str`
        The filed's name.
    type : `type`
        The binded class.
    fdel : `None`, `Any`
        Deleter function if applicable.
    fget : `None`, `Any`
        Getter function if applicable.
    fset : `None`, `Any`
        Setter function if applicable.
    """
    __slots__ = ('fdel', 'fget', 'fset')
    
    @copy_docs(ObjectBinderBase.__new__)
    def __new__(cls, name, type_):
        self = ObjectBinderBase.__new__(cls, name, type_)
        self.fdel = getattr(type_, '__delete__', None)
        self.fget = getattr(type_, '__get__')
        self.fset = getattr(type_, '__set__', None)
        return self
    
    
    @copy_docs(ObjectBinderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return self.fget(bind_object, instance, type_)
    
    
    @copy_docs(ObjectBinderBase.__get__)
    def __set__(self, instance, value):
        fset = self.fset
        if fset is None:
            raise AttributeError(f'Read only attribute: {self.name!r}.')
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return fset(bind_object, instance, value)
    
    
    @copy_docs(ObjectBinderBase.__delete__)
    def __delete__(self, instance):
        fdel = self.fdel
        if fdel is None:
            raise AttributeError(f'Cannot delete attribute: {self.name!r}.')
        
        try:
            bind_object = self.cache[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.cache[instance] = bind_object
        
        return fdel(bind_object, instance)



__all__ = ()

from .shared import convert_nothing


def convert_str__triggered_rule_name(name, data):
    return convert_nothing('trigger_rule_name', data)


AUTO_MODERATION_CONVERTERS = {
    'triggered_rule_name': convert_str__triggered_rule_name,
}



__all__ = ()

from ...permission import PermissionOverwriteTargetType


def detail_converter_channel_id(key, value):
    return 'channel_id', int(value)


def detail_converter_amount(key, value):
    return 'amount', int(value)


def detail_converter_days(key, value):
    return 'days', int(value)


def detail_converter_users_removed(key, value):
    return 'users_removed', int(value)


def detail_converter_message(key, value):
    return 'message_id', int(value)


def detail_converter_permission_overwrite_target_id(key, value):
    return 'target_id', int(value)


def detail_converter_permission_overwrite_target_type(key, value):
    if PermissionOverwriteTargetType.VALUE_TYPE is int:
        value = int(value)
    
    return 'target_type', PermissionOverwriteTargetType.get(value)


def detail_converter_nothing(key, value):
    return key, value


DETAIL_CONVERTERS = {
    'channel_id': detail_converter_channel_id,
    'count': detail_converter_amount,
    'delete_member_days': detail_converter_days,
    'id': detail_converter_permission_overwrite_target_id,
    'members_removed': detail_converter_users_removed,
    'message_id': detail_converter_message,
    'role_name': detail_converter_nothing,
    'type': detail_converter_permission_overwrite_target_type,
}



__all__ = ('ExtensionModuleProxyType',)

import warnings
from types import ModuleType

from scarletio import get_last_module_frame, include

from ..constants import EXTENSIONS

ExtensionModuleSpecType = include('ExtensionModuleSpecType')


class ExtensionModuleProxyType(ModuleType):
    """
    Represents an extension module's proxy.
    """
    def __init__(self, spec):
        """
        Creates a new extension module proxy.
        
        Parameters
        ----------
        spec : ``ExtensionModuleSpecType``
            Module specification.
        """
        ModuleType.__init__(self, spec.name)
        
        self.__spec__ = spec
        self.__file__ = spec.origin
    
    
    def __setattr__(self, attribute_name, attribute_value):
        ModuleType.__setattr__(self, attribute_name, attribute_value)
        
        if attribute_name in {
            '__spec__',
            '__loader__',
            '__package__',
            '__spec__',
            '__path__',
            '__file__',
            '__cached__',
        }:
            pass
        
        elif isinstance(attribute_value, ExtensionModuleProxyType):
            spec = self.__spec__
            
            module = spec.get_module()
            if (module is not None):
                setattr(module, attribute_name, attribute_value)
        
        else:
            warnings.warn(
                f'Unallowed attribute assignment: `{attribute_name} = {attribute_value!r}` of type '
                f'`{type(attribute_value).__name__}` to `{self.__spec__.name}`'
            )
    
    
    def __getattr__(self, attribute_name):
        spec = self.__spec__
        module = spec.get_module()
        
        if module is None:
            raise RuntimeError(
                f'Extension, {self.__extension__!r} is not yet initialized!'
            )
        
        attribute_value = getattr(module, attribute_name)
        
        frame = get_last_module_frame()
        if (frame is None):
            spec = None
        else:
            spec = frame.f_globals.get('__spec__', None)
        
        if spec is None:
            extension = None
        else:
            extension = EXTENSIONS.get(spec.name)
        
        current_extension = EXTENSIONS.get(spec.name, None)
        
        if (extension is not None) and (current_extension is not None) and (extension is not current_extension):
            extension.add_child_extension(current_extension)
            current_extension.add_parent_extension(extension)
        
        return attribute_value
    
    
    def __repr__(self):
        spec = getattr(self, '__spec__', None)
        
        repr_parts = ['<', type(self).__name__]
        
        if spec is None:
            repr_parts.append(' unknown location')
        
        else:
            repr_parts.append(' from ')
            repr_parts.append(spec.name)
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)



__all__ = ('EXTENSIONS',)

from scarletio import WeakValueDictionary


EXTENSIONS = WeakValueDictionary()

EXTENSION_STATE_UNDEFINED = 0
EXTENSION_STATE_LOADED = 1
EXTENSION_STATE_UNLOADED = 2
EXTENSION_STATE_UNSATISFIED = 3

EXTENSION_STATE_VALUE_TO_NAME = {
    EXTENSION_STATE_UNDEFINED: 'undefined',
    EXTENSION_STATE_LOADED: 'loaded',
    EXTENSION_STATE_UNLOADED: 'unloaded',
    EXTENSION_STATE_UNSATISFIED: 'unsatisfied',
}

LOADING_EXTENSIONS = set()

# Extension roots define core part for extensions.
EXTENSION_ROOTS = set()


ABSOLUTE_PATH_EXTENSION_NAME_PREFIX = '<extension>.'



__all__ = ('COMMAND_NAME_TO_COMMAND', 'COMMANDS',)

from os.path import dirname as get_directory_name, join as join_paths

from .. import __package__ as PACKAGE_NAME


COMMANDS = set()

COMMAND_NAME_TO_COMMAND = {}

COMMAND_FOLDER = join_paths(
    get_directory_name(__file__),
    'commands',
)

COMMAND_IMPORT_ROUTE = (PACKAGE_NAME, 'main', 'commands')
